/**
 * \file kakuro.h
 * \author Léo Perrin
 * \date 2012/01/24
 */
#ifndef _KAKURO_H_
#define _KAKURO_H_

#include <vector>
#include <iostream>


/**
 * \class Kakuro
 * \brief This class is used to manipulate kakuro grids.
 * 
 * It allows several operations such as parsing a file correctly
 * formatted to generate the kakuro instance or returning several
 * arrays and grids to be used by a program working on the kakuro.
 */
class Kakuro
{
	private:
                /**
                 * \brief The number of columns in the kakuro grid.
                 */
                unsigned int mSize;

                /**
                 * \brief The number of lines in the kakuro grid.
                 */
                unsigned int nSize;

                /**
                 * \brief The structure of the grid. Contains 0 if the
                 * cell cannot be filled, 1 otherwise. This matrix is
                 * of the same size as the kakuro. It can receive the
                 * values at each point of the grid.
                 */
                std::vector< std::vector<unsigned int> > gridStruct;
                
                /**
                 * \brief Contains zero if their is no information on
                 * horizontal sum, the value of the sum otherwise.
                 */
                std::vector< std::vector<unsigned int> > horizontalSum;
                
                /**
                 * \brief Contains zero if their is no information on
                 * vertical sum, the value of the sum otherwise.
                 */
                std::vector< std::vector<unsigned int> > verticalSum;

                /**
                 * \brief Contains the successive values of the sums
                 * given by the hints. The index of each hint matches
                 * that of the sumsAbscissa and sumsOrdinate
                 * attribute.
                 */
                std::vector<unsigned int> sumsValue;

                /**
                 * \brief Contains the successive lists of abscissas of
                 * the cells involved in each sum. The index of each
                 * hint matches that of the sumsValue and sumsOrdinate
                 * attribute.
                 */
                std::vector< std::vector<unsigned int> > sumsAbscissa;

                /**
                 * \brief Contains the successive lists of ordinates of
                 * the cells involved in each sum. The index of each
                 * hint matches that of the sumsValue and sumsAbscissa
                 * attribute.
                 */
                std::vector< std::vector<unsigned int> > sumsOrdinate;


	public:
                        
                /**
                 * \brief Studies a cell in order to see whether it
                 * should be filled in or contains a sum to reach.
                 *
                 * \param cell The string containing the content of the
                 * cell to parse.
                 * \param type Gets assigned to true if it is a free
                 * cell, false if we cannot write on it.
                 * \param horizontal Gets assigned to the value of the
                 * horizontal sum if any, 0 otherwise.
                 * \param vertical The same as vertical sum, except
                 * that it obviously concerns the vertical sum.
                 */
                static void parseCell(std::string cell,
                                       bool * type,
                                       unsigned int * horizontal,
                                       unsigned int * vertical );


                /**
                 * \brief Pretty prints the gridStruct, verticalSum
                 * and horizontalSum.
                 */
                void printGrids();
                

                /**
                 * \brief Reads a kakoru formatted stream and assigns
                 * the attributes of this instance accordingly. Once
                 * the call to this method is over, this instance
                 * contains a complete description of the kakuro grid.
                 *
                 * \param input The stream from which to read.
                 */
                void parseGrid(std::istream * input);

                /**
                 * \brief Pretty prints the kakuro grid in a file.
                 *
                 * \param out The stream to the file in which the grid
                 * should be printed.
                 * \param isEmpty If true then outputs '-' instead of
                 * '1' in the empty cells. Otherwise, pretty prints
                 * the complete grid.
                 */
                void printToStream(std::ostream&,bool);

                /**
                 * /brief Uses the grids gridStruct, horizontalSum and
                 * verticalSum to assign the sumsValue, sumsAbscissa
                 * and sumsOrdinate attributes.
                 */
                void grids2sums();

                /**
                 * \brief Sets the value at a given point by modifying
                 * the gridStruct attribute.
                 *
                 * \param j The index of the column of the cell to
                 * modify.
                 * \param i The index of the line of the cell to
                 * modify.
                 * \param value The value to assign to this cell.
                 */
                void assignValue(unsigned int, unsigned int, unsigned int);

                /**
                 * \brief Geter for nSize
                 * \return nSize.
                 */
                unsigned int getNSize();

                /**
                 * \brief Geter for mSize
                 * \return mSize.
                 */
                unsigned int getMSize();

                /**
                 * \brief Geter for the number of hints
                 * \return The size of sumsValue.
                 */
                unsigned int getNHints();

                /**
                 * \brief Geter for gridStruct.
                 * \return gridStruct.
                 */
                std::vector< std::vector<unsigned int> > getGridStruc();

                /**
                 * \brief Geter for horizontalSum.
                 * \return horizontalSum.
                 */
                std::vector< std::vector<unsigned int> > getHorizontalSum();

                /**
                 * \brief Geter for verticalSum.
                 * \return verticalSum.
                 */
                std::vector< std::vector<unsigned int> > getVerticalSum();

                /**
                 * \brief Geter for sumsValue.
                 * \return sumsValue.
                 */
                std::vector<unsigned int> getSumsValue();

                /**
                 * \brief Geter for sumsOrdinate.
                 * \return sumsOrdinate.
                 */
                std::vector< std::vector<unsigned int> > getSumsOrdinate();

                /**
                 * \brief Geter for sumsAbscissa.
                 * \return sumsAbscissa.
                 */
                std::vector< std::vector<unsigned int> > getSumsAbscissa();

//                ~Kakuro();
                };

#endif
