#include <stdlib.h>
#include <vector>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include <sstream>

#include "../cnflib/variableset.h"
#include "../cnflib/clause.h"
#include "../cnflib/formula.h"
#include "kakuro.h"




/**
 * \brief Generates all the subset of [1,9] of size nCells such that
 * the sum of their elements reaches up to wantedSum. In order not to
 * have a huge number of such subsets, we add the constraint that they
 * must be ordered. It is a recusrsive function.
 *
 * \param possibleSums Will contain all the possible ways to obtain
 * the correct sum it is modified recursively.
 * \param current The possible values of the cells we are currently
 * considering.
 * \param nCells The number of cells in the summation.
 * \param wantedSum Obviously, the sum we want to reach.
 */
void allSums(
        std::vector< std::vector<unsigned int> > * possibleSums,
        std::vector<unsigned int> * current,
        int nCells,
        int wantedSum )
{
        if (nCells == 0) // If we reached the last cell
        {
                if (wantedSum == 0)
                        possibleSums->push_back(*current);
                        // If we have the correct sum, we keep it
        }
        else
                for (unsigned int i=1; i<=9; i++)
                {
                        if ( 
                                ((int)(wantedSum-i) >= (nCells-1)*1)
                             && ((int)(wantedSum-i) <= (nCells-1)*9)
                             && ( 
                                     (current->size() == 0)
                                     || ( (*current)[current->size()-1] < i)
                                )
                            )
                        {
                                std::vector<unsigned int> * newVector
                                        = new std::vector<unsigned int>;
                                *newVector = *current;
                                newVector->push_back(i);
                                allSums(possibleSums,newVector,nCells-1,wantedSum-i);
                        }
                }
        delete current;
}


/**
 * \brief No two cells in the same hint shall have the same
 * value. This function adds to the Formula form the clauses modeling
 * these axioms.
 * 
 * \param form The formula to modify.
 * \param kak The kakuro grid to consider.
 */
void disjointClauses(Formula * form, Kakuro kak)
{
        std::vector< std::vector<unsigned int> >  sumsAbscissa = kak.getSumsAbscissa();
        std::vector< std::vector<unsigned int> >  sumsOrdinate = kak.getSumsOrdinate();
        for (unsigned int nHints=0; nHints<sumsAbscissa.size(); nHints++)
                for (unsigned int i1=0; i1<sumsAbscissa[nHints].size()-1; i1++)
                        for (unsigned int i2=i1+1; i2<sumsAbscissa[nHints].size(); i2++)
                                for (unsigned int value=1; value<=9; value++)
                                {
                                        form->newClause();
                                        form->addLit(0, false, sumsAbscissa[nHints][i1], sumsOrdinate[nHints][i1], value-1);
                                        form->addLit(0, false, sumsAbscissa[nHints][i2], sumsOrdinate[nHints][i2], value-1);
                                        form->pushClause();
                                }
}


/**
 * \brief We add \f$\bar{x_ijk} \land \bar{x_ijk'}\f$ for k<k'.
 *
 * \param i First coordinate of the cell.
 * \param j Second coordinate of the cell.
 * \param form The formula in which clauses will be added.
 */
void oneValuePerCell(unsigned int i, unsigned int j, Formula * form)
{
        for (unsigned int k1=0; k1<8; k1++)
		for (unsigned int k2=k1+1; k2<9; k2++)
		{
			form->newClause();
			form->addLit(0,false,i,j,k1);
			form->addLit(0,false,i,j,k2);
			form->pushClause();
		}
}


/**
 * \brief Adds to the Formula form the clauses modeling the fact that
 * each cell must have a value within a certain set. It is subroutine
 * of sumAxioms().
 *
 * \param form The Formula to modify.
 * \param nHints The rank of the hint to consider.
 * \param nValSets The rank of the possible subset of [1,9] to consider.
 * \param kak The kakuro grid concerned.
 * \param possibleSums The set of all the legal subsets of [1,9].
 */
void allowedValues(Formula * form,
                   unsigned int nHints,
                   unsigned int nValSets,
                   Kakuro kak,
                   std::vector< std::vector<unsigned int> > * possibleSums)
{
        std::vector< std::vector<unsigned int> > sumsAbscissa = kak.getSumsAbscissa();
        std::vector< std::vector<unsigned int> > sumsOrdinate = kak.getSumsOrdinate();
        for (unsigned int elem=0; elem<(*possibleSums)[nValSets].size(); elem++)
        {
                form->newClause();
                form->addLit(1,false,nHints,nValSets); // not y_{nHints,nValSets}
                for (unsigned int l=0; l<sumsAbscissa[nHints].size(); l++)
                        form->addLit(0,
                                     true,
                                     sumsAbscissa[nHints][l],
                                     sumsOrdinate[nHints][l],
                                     (*possibleSums)[nValSets][elem]-1);
                form->pushClause();
        }
}


/**
 * \brief Adds new closes to the formula, namely those stating that
 * the literals must have a value corresponding to the sum.
 *
 * It works by generating, for each hint indexed by nHints, all the
 * possible subsets of [1,9] of the correct size such that
 * sum(subset)=hint[nHints]. For each such valid subset (indexed by
 * nValSets), it creates a new variable y_{nHints,nValSets} which
 * implies that the each cell involved in this hint has a value in
 * that subset. For instance, y_{1,1} => x_{1,1,1} \/ x_{1,1,2}.
 *
 * \param form The formula to update.
 * \param kak The kakuro grid to consider.
 */
void sumAxioms(Formula * form, Kakuro kak)
{
        std::vector<unsigned int>  sumsValue = kak.getSumsValue();
        std::vector< std::vector<unsigned int> >  sumsAbscissa = kak.getSumsAbscissa();
        std::vector< std::vector<unsigned int> >  sumsOrdinate = kak.getSumsOrdinate();
        std::vector< std::vector<unsigned int> > *  possibleSums = new std::vector< std::vector<unsigned int> >;

        for (unsigned int nHints=0; nHints<kak.getNHints(); nHints++)
        {
                possibleSums->clear();
                allSums(possibleSums,
                        new std::vector<unsigned int>,
                        sumsAbscissa[nHints].size(),
                        sumsValue[nHints]);
                if (possibleSums->size() == 0)
                {
                        std::cout<<"ERROR: Kakuro grid impossible, no"
                                 <<"way to make "<<sumsValue[nHints]
                                 <<"with "<<sumsAbscissa[nHints].size()
                                 <<"cells!"<<std::endl;
                        exit(100);
                }
                else
                {
                        // non-emptiness
                        form->newClause();
                        for (unsigned int nValSets=0; nValSets<possibleSums->size(); nValSets++)
                                form->addLit(1,true,nHints,nValSets);
                        form->pushClause();
                        // non-duplicated legal subsets
                        for (unsigned int nValSets1=0; nValSets1<possibleSums->size()-1; nValSets1++)
                        {
                                form->newClause();
                                for (unsigned int nValSets2=nValSets1+1; nValSets2<possibleSums->size(); nValSets2++)
                                {
                                        form->addLit(1,false,nHints,nValSets1);
                                        form->addLit(1,false,nHints,nValSets2);
                                }
                                form->pushClause();
                        }
                        // correct sums
                        for (unsigned int nValSets=0; nValSets<possibleSums->size(); nValSets++)
                        allowedValues(form, nHints, nValSets, kak, possibleSums);
                }
        }
}




/**
 * \brief Main function of this program.
 *
 * Parses the input we want to read a kakuro instance. It then turns it
 * into a CNF formula using a Formula instance from the cnflib.
 */
int main(int argc, char ** argv)
{
        Kakuro kak;
        // initialising kakuro
        if (argc==1)
	{
                kak.parseGrid(&std::cin);
        }
	else
	{
        	std::string name (argv[1]);
		std::ifstream input(name.c_str());
                kak.parseGrid(&input);
		input.close();
        }
        kak.grids2sums();
        // copying kakuro grid
        std::string currentKakuro("currentKakuro.txt");
        std::ofstream out(currentKakuro.c_str());
        kak.printToStream(out,true);
        out.close();
        // initialising the variable set
        VariableSet vars(kak.getNSize(), kak.getMSize(), 9);
        vars.addVars(kak.getNHints(),12);
        Formula * form = new Formula(&vars);
        // adding clauses to the formula
        disjointClauses(form,kak);
        for (unsigned int i=0; i<kak.getNSize(); i++)
                for (unsigned int j=0; j<kak.getMSize(); j++)
                        oneValuePerCell(i,j,form);
        sumAxioms(form,kak);
        // outputing the formula to stdout
        std::string comment ("Kakuro as a CNF formula");
        form->toFile(&std::cout,comment);
	std::cout<<std::endl;
	return 0;
}
